
import re
import json
import scrapy
from urllib.parse import urlencode
from scrapy.loader import ItemLoader
from datetime import datetime
import pytz


def clean_text(text):

    if text is None:

        return '-'

    if text is not None:

        cleaned_value = re.sub(r'<(?!br\s?\/?)[^>]*>', '', text)
                
        # Remove leading and trailing whitespaces, including newline characters
        cleaned_value = cleaned_value.strip()

        # Replace consecutive newline characters with a single space
        cleaned_value = re.sub(r'\n+', ' ', cleaned_value)

        # Replace other unwanted characters
        cleaned_value = cleaned_value.replace("\xa0", "").replace("  ","").replace("m<sup>2</sup>","").replace("m<sup>3</sup>","").replace("\r","").replace("\t","").replace("<span>","").replace("</span>","").strip()

        # Use strip() to remove leading and trailing whitespace and newline characters
        return cleaned_value
    
class IndeedJobSpider(scrapy.Spider):
  name = "indeed_jobs"

  def get_indeed_search_url(self, keyword, location, offset=0, radius =50,fromage=14):
      parameters = {
          "q": keyword,
          "l": location,
          "filter": 0,
          "start": offset,
          "sort": "date",
          'radius': radius,
          "fromage": fromage  # Added fromage parameter (search jobs from the last n days)
      }
      return "https://nl.indeed.com/jobs?" + urlencode(parameters)

  def start_requests(self):
    #parameters sent to the proxy api, fromage means the number of days that the ad has been placed on indee, and radius means the number of kilometers from the search word in location list.
    keyword_list = ['python']
    location_list = ['nederland']
    fromage = 14
    radius = 50

    for keyword in keyword_list:
      for location in location_list:
        indeed_jobs_url = self.get_indeed_search_url(keyword, location,radius,fromage)
        yield scrapy.Request(url=indeed_jobs_url,
                             callback=self.parse_search_results,
                             meta={
                                'keyword': keyword,
                                'location': location,
                                'offset': 0,
                                'sort': 'date',
                                'radius': radius,
                                'fromage': fromage

                             })

  def parse_search_results(self, response):
    script_tag = re.findall(
        r'window.mosaic.providerData\["mosaic-provider-jobcards"\]=(\{.+?\});',
        response.text)
    if script_tag:
      json_blob = json.loads(script_tag[0])

      if response.meta['offset'] == 0:
        meta_data = json_blob["metaData"]["mosaicProviderJobCardsModel"][
            "tierSummaries"]
        num_results = sum(category["jobCount"] for category in meta_data)
        if num_results > 1000:
          num_results = 50

        for offset in range(10, 20, 10):
          url = self.get_indeed_search_url(response.meta['keyword'],
                                           response.meta['location'], 
                                           offset,
                                           response.meta['radius'],
                                           response.meta['fromage'])
          yield scrapy.Request(url=url,
                               callback=self.parse_search_results,
                               meta={
                                   'keyword': response.meta['keyword'],
                                   'location': response.meta['location'],
                                   'offset': offset,
                                   'sort':'date',
                                   'radius': response.meta['radius'],
                                   'fromage': response.meta['fromage']
                               })

      jobs_list = json_blob['metaData']['mosaicProviderJobCardsModel'][
          'results']
      for index, job in enumerate(jobs_list):
        if job.get('jobkey') is not None:
          job_url = 'https://nl.indeed.com/viewjob?viewtype=embedded&jk=' + job.get(
              'jobkey')
          yield scrapy.Request(url=job_url,
                               callback=self.parse_job,
                               meta={
                                'keyword': response.meta['keyword'],
                                'location': response.meta['location'],
                                'offset': response.meta['offset'],
                                'position': index,
                                'jobKey': job.get('jobkey'),
                                'company': job.get('company'),
                                'maxSalary': job.get('extractedSalary')['max'] if job.get('extractedSalary') and job.get('extractedSalary')['max'] > 0 else 0,
                                'minSalary': job.get('extractedSalary')['min'] if job.get('extractedSalary') and job.get('extractedSalary')['min'] > 0 else 0,
                                'salaryType': job.get('extractedSalary')['type'] if job.get('extractedSalary') is not None else 'none',
                                'pubDate': job.get('pubDate'),
                                'jobTitle': job.get('title'),
                                'jobLocationCity': job.get('jobLocationCity'),
                                'jobLocationPostal': job.get('jobLocationPostal'),
                                'jobLocationState': job.get('jobLocationState')
                               })

  def parse_job(self, response):
    script_tag = re.findall(r"_initialData=(\{.+?\});", response.text)
    if script_tag:
      job_info = {}
      json_blob = json.loads(script_tag[0])
      job = json_blob["jobInfoWrapperModel"]["jobInfoModel"]
      job_info['source'] = 'indeed'
      job_info['job_key'] = response.meta['jobKey']
      job_info['url'] = response.url
      job_info['salary'] = {
        "max":response.meta['maxSalary'],
        "min":response.meta['minSalary'],
        "type":response.meta['salaryType']
      }
      job_info['title'] = response.meta['jobTitle']
      job_info['place'] = response.meta['jobLocationCity']
      job_info['place_postal'] = response.meta['jobLocationPostal']
      job_info['place_full'] = job['jobInfoHeaderModel']['formattedLocation']
      job_info['recruiter'] = {
        'name':response.meta['company']
      }
      job_info['benefits'] = response.xpath('//div[@class="css-1oelwk6 eu4oa1w0"]/div[@class="css-k3ey05 eu4oa1w0"]//li/text()').getall()
      job_info['description'] = clean_text(job.get('sanitizedJobDescription',''))
      job_info['hours'] = {
        "type":job['jobMetadataHeaderModel']['jobType']
      }

      timestamp_ms = response.meta['pubDate']

      # Convert to seconds by dividing by 1000
      timestamp_s = timestamp_ms / 1000
      
      # Define the Amsterdam timezone
      amsterdam_tz = pytz.timezone('Europe/Amsterdam')

      # Convert timestamp to datetime in Amsterdam timezone
      dt = datetime.fromtimestamp(timestamp_s, tz=pytz.utc).astimezone(amsterdam_tz)

      job_info['dateRefreshed'] = dt.isoformat()

      yield job_info



import scrapy
import json
import re
from datetime import datetime
import numpy as np
class AsunnotrentapispiderSpider(scrapy.Spider):

    name = "asunnotrentapispider"
    base_url = "https://asunnot.oikotie.fi/vuokra-asunnot?pagination={}&cardType=101"

    def __init__(self, page_limit=10, stored_newest_timestamp=0, *args, **kwargs):
        super(AsunnotrentapispiderSpider, self).__init__(*args, **kwargs)
        self.page_limit = int(page_limit)
        self.stored_newest_timestamp = int(stored_newest_timestamp) if stored_newest_timestamp else 0
        self.current_page = 1
        self.start_urls = [self.base_url.format(self.current_page)]

    def parse(self, response):
        # Extract necessary tokens from the page content
        ota_cuid = response.xpath('//meta[@name="cuid"]/@content').get()
        ota_token = response.xpath('//meta[@name="api-token"]/@content').get()
        ota_loaded = response.xpath('//meta[@name="loaded"]/@content').get()

        # Extract cookies
        cookies = response.headers.getlist('Set-Cookie')
        cookies_dict = {}
        for cookie in cookies:
            cookie_parts = cookie.decode().split(';')[0].split('=')
            cookies_dict[cookie_parts[0]] = cookie_parts[1]
        url = response.url

        # Prepare headers for the API request
        headers = {
            'Accept': 'application/json',
            'Accept-Encoding': 'gzip, deflate, br, zstd',
            'Accept-Language': 'en,zh-CN;q=0.9,zh;q=0.8',
            'Cache-Control': 'no-cache',
            'Ota-Cuid': ota_cuid,
            'Ota-Loaded': ota_loaded,
            'Ota-Token': ota_token,
            'Pragma': 'no-cache',
            'Priority': 'u=1, i',
            'Referer': url,
            'Sec-Ch-Ua': '"Chromium";v="124", "Google Chrome";v="124", "Not-A.Brand";v="99"',
            'Sec-Ch-Ua-Mobile': '?0',
            'Sec-Ch-Ua-Model': '""',
            'Sec-Ch-Ua-Platform': '"Windows"',
            'Sec-Ch-Ua-Platform-Version': '"15.0.0"',
            'Sec-Fetch-Dest': 'empty',
            'Sec-Fetch-Mode': 'cors',
            'Sec-Fetch-Site': 'same-origin',
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36',
        }

        offset = (self.current_page - 1) * 24

        api_url = f'https://asunnot.oikotie.fi/api/search?cardType=101&limit=24&offset={offset}&sortBy=published_sort_desc'

        yield scrapy.Request(url=api_url, headers=headers, cookies=cookies_dict, callback=self.parse_api, cb_kwargs={"url":url})

    def parse_api(self, response, url= None):

        data = response.json()

        # Filter entries where 'listingType' is 1
        found_entries = [entry for entry in data['cards']]

        if not found_entries:
            return

        # Extract the 'published' dates from the filtered entries
        published_dates = [entry['meta']['published'] for entry in found_entries]

        # Convert the dates to datetime objects
        published_dates = [datetime.strptime(date, '%Y-%m-%d %H:%M:%S') for date in published_dates]

        # Convert datetime objects to timestamps
        published_timestamps = [int(date.timestamp()) for date in published_dates]

        # Find the oldest (maximum) timestamp
        newest_timestamp = max(published_timestamps)

        # Flag to indicate if the loop should be broken
        should_break = False

        if newest_timestamp <= self.stored_newest_timestamp:
            should_break = True

        if should_break == True:
            print('This is the last page.')
            return
        
        for item in data['cards']:  # Adjust this according to the actual JSON structure
            link = item.get('url')
            
            headers ={
                'referer' : url
            }

            published_timestamp = int((datetime.strptime(item['meta']['published'], '%Y-%m-%d %H:%M:%S')).timestamp())
            
            # Check the condition only for items with listingType == 1
            if published_timestamp <= self.stored_newest_timestamp:
                continue
                    
            yield scrapy.Request(url = link, callback = self.parse_property,headers=headers)

        # Handle pagination
        if self.current_page < self.page_limit:
            self.current_page += 1
            next_page_url = self.base_url.format(self.current_page)
            yield scrapy.Request(url=next_page_url, callback=self.parse)

    def parse_property(self,response):

        property_info = {}
        
        property_info['source'] = response.url

        if 'vuokra' in property_info['source']:

            property_info['property_listing_type'] = 'rent'

        else:

            property_info['property_listing_type'] = 'sale'

        property_info['property_is_new_construction'] = False

        try:

            property_info['property_description'] = " ".join(response.css('div.listing-overview p::text').getall())
        except:

            pass
        
        try:
            property_info['property_floor_number'] = re.search(r'\d+', response.xpath('//dt[text()="Kerros"]/following-sibling::dd/text()').get()).group() if response.xpath('//dt[text()="Kerros"]/following-sibling::dd/text()').get() else None
        
        except:
            pass

        try:
            property_info['property_configuration'] = response.css('dt:contains("Huoneiston kokoonpano") + dd::text').get()

        except:
            pass

        try:
            property_info['property_condition_finnish'] = response.css('dt:contains("Kunto") + dd::text').get()

        except:
            pass

        try:
            property_info['property_kitchen_features'] = response.css('dt:contains("Keittiön varusteet") + dd::text').get()
        
        except:
            pass

        try:
            property_info['property_bathroom_features'] = response.css('dt:contains("Kylpyhuoneen varusteet") + dd::text').get()

        except:
            pass

        try:
            property_info['property_renovation_info'] = response.css('dt:contains("Tehdyt remontit") + dd::text').get()

        except:
            pass

        try:
            property_info['property_additional_terms'] = response.css('dt:contains("Muut ehdot") + dd::text').get()
        except:
            pass

        try:
            property_info['property_sauna'] = response.css('dt:contains("Taloyhtiössä on sauna") + dd::text').get()

        except:
            pass

        try:
            property_info['property_heating'] = response.css('dt:contains("Lisätietoja lämmityksestä") + dd::text').get()

        except:
            pass
        
        try:
            property_info['property_energy_label'] = response.css('dt:contains("Energialuokka") + dd::text').get()

        except:
            pass

        try:
            image_urls = response.css('div.tabs-content a::attr(href)').getall()
            filtered_image_urls = [url for url in image_urls if url.startswith('https://cdn.asunnot.oikotie.fi/')]

            property_info['media_set'] = filtered_image_urls

        except:
            pass

        # Iterate through each script tag
        for script_text in response.css("script::text").getall():
            # Check if the script text contains 'analytics'
            if 'analytics' in script_text:
                # Extract JSON data from the script
                json_data = script_text.split("window.page=")[0].strip().rstrip(';')[script_text.split("window.page=")[0].strip().rstrip(';').find('{'):]
                json_data = json.loads(json_data)
                
                # Populate property_info dictionary with extracted data
                property_info['property_housing_type_finnish'] = json_data['analytics']['apartmentType']
                property_info['address_full'] = json_data['address']
                match = re.match(r'^(.*?)(\d+)(.*)\,', property_info['address_full'])
                if match:
                    property_info['address_street'] = match.group(1).strip()
                    property_info['address_house_number'] = match.group(2).strip()
                    property_info['address_addition'] = match.group(3).strip()
                    property_info['address_city'] = re.findall(r'\b\w+\b', property_info['address_full'])[-1].strip()
                else:
                    property_info['address_street'] = re.findall(r'^[a-zA-Z]+', property_info['address_full'])[0].strip()
                    property_info['address_city'] = re.findall(r'\b\w+\b', property_info['address_full'])[-1].strip()
                property_info['property_living_area_sqm'] = json_data['analytics']['size']
                # Define the bins and labels for categorization
                area = property_info["property_living_area_sqm"]
                if not area:
                    property_info['bucket'] = np.NaN
                elif area < 25:
                    property_info['bucket'] = 1
                elif area < 50:
                    property_info['bucket'] = 2
                elif area < 75:
                    property_info['bucket'] = 3
                elif area < 100:
                    property_info['bucket'] = 4
                elif area < 150:
                    property_info['bucket'] = 5
                else:
                    property_info['bucket'] = 6
                property_info['property_rent_price'] = json_data['analytics']['price']
                property_info['property_rent_price_currency'] = 'EUR'
                try:
                    property_info['property_rent_price_per_sqm'] = property_info['property_rent_price'] / property_info['property_living_area_sqm']
                except:
                    pass
                property_info['timeline_advertisement_creation'] = json_data['analytics']['published']
                try:

                    property_info['address_postal_code'] = json_data['analytics']['zipCode']

                except KeyError:

                    property_info['address_postal_code'] = None
                    
                property_info['property_construction_year'] = json_data['analytics']['apartmentBuildYear']
                
                # Break the loop after finding the relevant script
                break

        # Iterate through each script tag again to find additional data
        for script_text in response.css("script::text").getall():
            # Check if the script text contains additional data (assuming it has 'numberOfRooms' and 'geo' keys)
            if 'numberOfRooms' in script_text and 'geo' in script_text:
                # Extract JSON data from the script
                json_data_1 = json.loads(script_text)
                
                # Populate property_info dictionary with additional data
                property_info['property_rooms_amount'] = json_data_1['numberOfRooms']
                property_info['latitude'] = json_data_1['geo']['latitude']
                property_info['longitude'] = json_data_1['geo']['longitude']
                property_info['address_country'] = json_data_1['address']['addressCountry']
                try:
                    property_info['address_municipality'] = json_data_1['address']['addressRegion']
                except:
                    pass

                try:
                    property_info['address_place'] = json_data_1['address']['addressLocality']
                except:
                    pass
                # Break the loop after finding the relevant script
                break

        yield property_info



import scrapy
import re
import xml.etree.ElementTree as ET
import json
from scrapy.selector import Selector
import numpy as np
class AsunnotspiderSpider(scrapy.Spider):
    name = "asunnotspider"
    start_urls = ["https://asunnot.oikotie.fi/sitemaps/index.xml"]

    def parse(self, response):

        # Preprocess the XML content to remove incorrect URLs
        xml_content = response.body.decode('utf-8')
        xml_content_cleaned = re.sub(r'<loc>https://asunnot.oikotie.fi/sitemapssunnot.oikotie.fi/', '<loc>https://asunnot.oikotie.fi/', xml_content)

        # Parse the cleaned XML content
        root = ET.fromstring(xml_content_cleaned)

        # Extract URLs from the cleaned XML
        sitemap_urls = [elem.text for elem in root.findall(".//{http://www.sitemaps.org/schemas/sitemap/0.9}loc")]

        # Filter URLs that start with 'https://asunnot.oikotie.fi/sitemaps/sm_ad'
        filtered_urls = [url for url in sitemap_urls if url.startswith('https://asunnot.oikotie.fi/sitemaps/sm_ad')]

        # Extract the content between <loc> tags
        for sitemap_url in filtered_urls:

            link = sitemap_url.strip()  # Remove any leading or trailing whitespace

            yield scrapy.Request(url = link, callback = self.get_urls)
    
    def get_urls(self,response):

        # Decode the response body and create a new Selector
        body = response.body.decode('utf-8')
        selector = Selector(text=body)

        # Extract URLs between <loc> tags that contain "myytavat-asunnot"
        urls = selector.xpath('//url/loc[contains(text(), "vuokra-asunnot")]/text()').getall()
        
        # Process each URL as needed
        for link in urls:
            yield scrapy.Request(url=link, callback=self.parse_property)
    
    def parse_property(self,response):

        property_info = {}
        
        property_info['source'] = response.url

        if 'vuokra' in property_info['source']:

            property_info['property_listing_type'] = 'rent'

        else:

            property_info['property_listing_type'] = 'sale'
        
        property_info['property_is_new_construction'] = False

        try:

            property_info['property_description'] = " ".join(response.css('div.listing-overview p::text').getall())
        except:

            pass
        
        try:
            property_info['property_floor_number'] = re.search(r'\d+', response.xpath('//dt[text()="Kerros"]/following-sibling::dd/text()').get()).group() if response.xpath('//dt[text()="Kerros"]/following-sibling::dd/text()').get() else None
        
        except:
            pass

        try:
            property_info['property_configuration'] = response.css('dt:contains("Huoneiston kokoonpano") + dd::text').get()

        except:
            pass

        try:
            property_info['property_condition_finnish'] = response.css('dt:contains("Kunto") + dd::text').get()

        except:
            pass

        try:
            property_info['property_kitchen_features'] = response.css('dt:contains("Keittiön varusteet") + dd::text').get()
        
        except:
            pass

        try:
            property_info['property_bathroom_features'] = response.css('dt:contains("Kylpyhuoneen varusteet") + dd::text').get()

        except:
            pass

        try:
            property_info['property_renovation_info'] = response.css('dt:contains("Tehdyt remontit") + dd::text').get()

        except:
            pass

        try:
            property_info['property_additional_terms'] = response.css('dt:contains("Muut ehdot") + dd::text').get()
        except:
            pass

        try:
            property_info['property_sauna'] = response.css('dt:contains("Taloyhtiössä on sauna") + dd::text').get()

        except:
            pass

        try:
            property_info['property_heating'] = response.css('dt:contains("Lisätietoja lämmityksestä") + dd::text').get()

        except:
            pass
        
        try:
            property_info['property_energy_label'] = response.css('dt:contains("Energialuokka") + dd::text').get()

        except:
            pass

        try:
            image_urls = response.css('div.tabs-content a::attr(href)').getall()
            filtered_image_urls = [url for url in image_urls if url.startswith('https://cdn.asunnot.oikotie.fi/')]

            property_info['media_set'] = filtered_image_urls

        except:
            pass

        # Iterate through each script tag
        for script_text in response.css("script::text").getall():
            # Check if the script text contains 'analytics'
            if 'analytics' in script_text:
                # Extract JSON data from the script
                json_data = script_text.split("window.page=")[0].strip().rstrip(';')[script_text.split("window.page=")[0].strip().rstrip(';').find('{'):]
                json_data = json.loads(json_data)
                
                # Populate property_info dictionary with extracted data
                property_info['property_housing_type_finnish'] = json_data['analytics']['apartmentType']
                property_info['address_full'] = json_data['address']
                match = re.match(r'^(.*?)(\d+)(.*)\,', property_info['address_full'])
                if match:
                    property_info['address_street'] = match.group(1).strip()
                    property_info['address_house_number'] = match.group(2).strip()
                    property_info['address_addition'] = match.group(3).strip()
                    property_info['address_city'] = re.findall(r'\b\w+\b', property_info['address_full'])[-1].strip()
                else:
                    property_info['address_street'] = re.findall(r'^[a-zA-Z]+', property_info['address_full'])[0].strip()
                    property_info['address_city'] = re.findall(r'\b\w+\b', property_info['address_full'])[-1].strip()
                property_info['property_living_area_sqm'] = json_data['analytics']['size']
                area = property_info["property_living_area_sqm"]
                if not area:
                    property_info['bucket'] = np.NaN
                elif area < 25:
                    property_info['bucket'] = 1
                elif area < 50:
                    property_info['bucket'] = 2
                elif area < 75:
                    property_info['bucket'] = 3
                elif area < 100:
                    property_info['bucket'] = 4
                elif area < 150:
                    property_info['bucket'] = 5
                else:
                    property_info['bucket'] = 6
                property_info['property_rent_price'] = json_data['analytics']['price']
                property_info['property_rent_price_currency'] = 'EUR'
                try:
                    property_info['property_rent_price_per_sqm'] = property_info['property_rent_price'] / property_info['property_living_area_sqm']
                except:
                    pass
                property_info['timeline_advertisement_creation'] = json_data['analytics']['published']
                try:

                    property_info['address_postal_code'] = json_data['analytics']['zipCode']

                except KeyError:

                    property_info['address_postal_code'] = None
                    
                property_info['property_construction_year'] = json_data['analytics']['apartmentBuildYear']
                
                # Break the loop after finding the relevant script
                break

        # Iterate through each script tag again to find additional data
        for script_text in response.css("script::text").getall():
            # Check if the script text contains additional data (assuming it has 'numberOfRooms' and 'geo' keys)
            if 'numberOfRooms' in script_text and 'geo' in script_text:
                # Extract JSON data from the script
                json_data_1 = json.loads(script_text)
                
                # Populate property_info dictionary with additional data
                property_info['property_rooms_amount'] = json_data_1['numberOfRooms']
                property_info['latitude'] = json_data_1['geo']['latitude']
                property_info['longitude'] = json_data_1['geo']['longitude']
                property_info['address_country'] = json_data_1['address']['addressCountry']
                try:
                    property_info['address_municipality'] = json_data_1['address']['addressRegion']
                except:
                    pass

                try:
                    property_info['address_place'] = json_data_1['address']['addressLocality']
                except:
                    pass
                # Break the loop after finding the relevant script
                break

        yield property_info


import re
import json
import scrapy
from urllib.parse import urlencode

class IndeedSearchSpider(scrapy.Spider):
    name = "indeed_search"
    custom_settings = {
        'FEEDS': { 'data/%(name)s_%(time)s.csv': { 'format': 'csv',}}
        }

    def get_indeed_search_url(self, keyword, location, offset=0):
        parameters = {"q": keyword, "l": location, "filter": 0, "start": offset}
        return "https://nl.indeed.com/jobs?" + urlencode(parameters)

    def start_requests(self):
        keyword_list = ['cyber']
        location_list = ['amsterdam']
        for keyword in keyword_list:
            for location in location_list:
                indeed_jobs_url = self.get_indeed_search_url(keyword, location)
                yield scrapy.Request(url=indeed_jobs_url, callback=self.parse_search_results, meta={'keyword': keyword, 'location': location, 'offset': 0})

    def parse_search_results(self, response):
        location = response.meta['location']
        keyword = response.meta['keyword'] 
        offset = response.meta['offset'] 
        script_tag  = re.findall(r'window.mosaic.providerData\["mosaic-provider-jobcards"\]=(\{.+?\});', response.text)
        if script_tag is not None:
            json_blob = json.loads(script_tag[0])

            ## Extract Jobs From Search Page
            jobs_list = json_blob['metaData']['mosaicProviderJobCardsModel']['results']
            for index, job in enumerate(jobs_list):
                yield {
                    'keyword': keyword,
                    'location': location,
                    'page': round(offset / 10) + 1 if offset > 0 else 1,
                    'position': index,
                    'company': job.get('company'),
                    'companyRating': job.get('companyRating'),
                    'companyReviewCount': job.get('companyReviewCount'),
                    'companyRating': job.get('companyRating'),
                    'highlyRatedEmployer': job.get('highlyRatedEmployer'),
                    'jobkey': job.get('jobkey'),
                    'jobTitle': job.get('title'),
                    'jobLocationCity': job.get('jobLocationCity'),
                    'jobLocationPostal': job.get('jobLocationPostal'),
                    'jobLocationState': job.get('jobLocationState'),
                    'maxSalary': job.get('extractedSalary')['max'] if job.get('extractedSalary') is not None else 0,
                    'minSalary': job.get('extractedSalary')['min'] if job.get('extractedSalary') is not None else 0,
                    'salaryType': job.get('extractedSalary')['type'] if job.get('extractedSalary') is not None else 'none',
                    'pubDate': job.get('pubDate')
                }
            
            ## Paginate Through Jobs Pages
            if offset == 0:
                meta_data = json_blob["metaData"]["mosaicProviderJobCardsModel"]["tierSummaries"]
                num_results = sum(category["jobCount"] for category in meta_data)

                # Limit to a maximum of 4 pages (offsets: 0, 10, 20, 30)
                max_offset = 30

                # if num_results > 1000:
                #     num_results = 50
                
                # for offset in range(10, num_results + 10, 10):

                for offset in range(10, max_offset + 10, 10):

                    url = self.get_indeed_search_url(keyword, location, offset)
                    yield scrapy.Request(url=url, callback=self.parse_search_results, meta={'keyword': keyword, 'location': location, 'offset': offset})







import scrapy
from scrapy.exceptions import CloseSpider
import re
import time
import json
import numpy as np
def clean_text(text):

    if text is None:

        return '-'

    if text is not None:
                
        # Remove leading and trailing whitespaces, including newline characters
        cleaned_value = text.strip()

        # Replace consecutive newline characters with a single space
        cleaned_value = re.sub(r'\n+', ' ', cleaned_value)

        # Replace other unwanted characters
        cleaned_value = cleaned_value.replace("\u2022", "").replace("\u2013", "").replace("\u2028", "").replace("\u0000", "").replace("\u00A0", "").strip()


        # Use strip() to remove leading and trailing whitespace and newline characters
        return cleaned_value

def normalize_key (text):

    if text is None:
        return ''
    
    if text is not None:
        

        normalized_key = text.replace('(p.m.)','').replace('(s)','').strip()

        # Replace spaces with underscores
        normalized_key = normalized_key.replace(' ', '_').replace('.','').strip()
    
        # Remove special characters
        normalized_key = ''.join(char for char in normalized_key if char.isalnum() or char in ['_', '-'])

        # Decapitalize the key
        normalized_key = normalized_key.lower()

        return normalized_key



class boligportalspiderSpider(scrapy.Spider):

    name = "boligportalspider"
    start_urls = ["https://www.boligportal.dk/en/rental-properties/?offset=0"]
    base_url = "https://www.boligportal.dk/en/rental-properties/?offset={}"
    
    all_possible_property_keys = set()

    def parse(self, response):

        # Extract the current page number from the URL
        current_page = int(response.url.split("offset=")[1])

        # Calculate the next page number with an increment of 18
        next_page_number = current_page + 18

        # Construct the next page URL
        next_page_url = self.base_url.format(next_page_number)
        
        properties_data = "".join(response.css("script::text")[-2].get())

        properties_data = json.loads(properties_data)

        properties = properties_data["props"]["page_props"]["results"]

        if not properties:

            return

        for property in properties:

            property_link = 'https://www.boligportal.dk' + property['url']

            yield scrapy.Request(url = property_link, callback = self.parse_property_page)

        yield scrapy.Request(next_page_url, callback=self.parse)

    
    def parse_property_page(self,response, property_info= None, retry_attempt=0):

        property_info = {}

        property_data = "".join(response.css("script::text")[-2].get())

        property_data = json.loads(property_data)

        data = property_data['props']['page_props']['ad']

        property_info['source'] = response.url

        property_info['address_country'] = 'Denmark'

        property_info['property_listing_type'] = 'rent'

        try:
            property_info['timeline_advertisement_creation'] = data['advertised_date']
        except :
            property_info['timeline_advertisement_creation'] = None

        try:
            property_info['address_city'] = data['city']
        except :
            property_info['address_city'] = None

        try:
            property_info['address_city_area'] = data['city_area']
        except :
            property_info['address_city_area'] = None

        try:
            property_info['address_street'] = data['street_name']
        except :
            property_info['address_street'] = None

        try:
            property_info['address_postal_code'] = str(data['postal_code'])
        except :
            property_info['address_postal_code'] = None

        try:
            property_info['property_description'] = data['description']
        except :
            property_info['property_description'] = None

        try:
            property_info['property_type'] = data['category']
        except :
            property_info['property_type'] = None

        try:
            property_info['property_advertisement_title'] = data['title']
        except :
            property_info['property_advertisement_title'] = None

        try:
            property_info['property_rooms_amount'] = int(data['rooms'])
        except :
            property_info['property_rooms_amount'] = np.nan

        try:
            property_info['property_living_area_sqm'] = data['size_m2']
        except :
            property_info['property_living_area_sqm'] = None
        # Define the bins and labels for categorization
        area = property_info["property_living_area_sqm"]
        if not area:
            property_info['bucket'] = np.NaN
        elif area < 25:
            property_info['bucket'] = 1
        elif area < 50:
            property_info['bucket'] = 2
        elif area < 75:
            property_info['bucket'] = 3
        elif area < 100:
            property_info['bucket'] = 4
        elif area < 150:
            property_info['bucket'] = 5
        else:
            property_info['bucket'] = 6

        try:
            property_info['property_rent_price'] = data['monthly_rent']
        except :
            property_info['property_rent_price'] = None

        try:
            property_info['property_rent_price_per_sqm'] = property_info['property_rent_price'] / property_info['property_living_area_sqm']
        except:
            pass
        try:
            property_info['property_rent_currency'] = data['monthly_rent_currency']
        except :
            property_info['property_rent_currency'] = None

        try:
            property_info['property_rent_monthly_extra_cost'] = data['monthly_rent_extra_costs']
        except :
            property_info['property_rent_monthly_extra_cost'] = None

        try:
            property_info['property_prepaid_rent'] = data['prepaid_rent']
        except :
            property_info['property_prepaid_rent'] = None

        try:
            property_info['property_deposit'] = data['deposit']
        except :
            property_info['property_deposit'] = None

        try:
            property_info['latitude'] = data['location']['lat']
        except :
            property_info['latitude'] = None

        try:
            property_info['longitude'] = data['location']['lng']
        except :
            property_info['longitude'] = None

        try:
            property_info['media_set'] = [image['url'] for image in data['images'][:15]]
        except :
            property_info['media_set'] = []

        try:
            property_info['property_floor_number'] = data['floor']
        except :
            property_info['property_floor_number'] = None

        try:
            property_info['property_rental_period'] = data['rental_period']
        except :
            property_info['property_rental_period'] = None

        try:
            property_info['address_neighborhood'] = data['city_level_3']
        except :
            property_info['address_neighborhood'] = None

        try:
            match = re.match(r'([A-Z])(\d{4})?', data['energy_rating'])
            if match:
                property_info['property_energy_label'] = match.group(1)
                property_info['property_energy_label_year'] = int(match.group(2)) if match.group(2) else None
        except :
            property_info['property_energy_label'] = None

        try:
            property_info['property_is_social_housing'] = data['social_housing']
        except :
            property_info['property_is_social_housing'] = None

        try:
            property_info['property_features_dictionary'] = [{feature: value} for feature, value in data['features'].items() if value]
        except :
            property_info['property_features_dictionary'] = None

        yield property_info





import scrapy
import re
import json
import numpy as np
import scrapy.resolver

def normalize_key (text):

    if text is None:
        return ''
    
    if text is not None:
        
        normalized_key = text.replace('(p.m.)','').replace('(s)','').strip()

        # Replace spaces with underscores
        normalized_key = normalized_key.replace(' ', '_').replace('.','').replace('-','_').strip()
    
        # Remove special characters
        normalized_key = ''.join(char for char in normalized_key if char.isalnum() or char in ['_', '-'])

        normalized_key = normalized_key.replace('__',"_")

        # Decapitalize the key
        normalized_key = normalized_key.lower()

        return normalized_key

def clean_number(number):

    if number is None:
        return ""
    
    cleaned_number = number.replace("€","").replace(",","").replace('','').strip()

    return cleaned_number

def clean_text(text):

    if text is None:

        return '-'

    if text is not None:

        cleaned_value = re.sub(r'<(?!br\s?\/?)[^>]*>', '', text)
                
        # Remove leading and trailing whitespaces, including newline characters
        cleaned_value = cleaned_value.strip()

        # Replace consecutive newline characters with a single space
        cleaned_value = re.sub(r'\n+', ' ', cleaned_value)

        # Replace other unwanted characters
        cleaned_value = cleaned_value.replace("\xa0", "").replace("  ","").replace("m<sup>2</sup>","").replace("m<sup>3</sup>","").replace("\r","").replace("\t","").replace("<span>","").replace("</span>","").strip()

        # Use strip() to remove leading and trailing whitespace and newline characters
        return cleaned_value

class BoligsidensellingspiderSpider(scrapy.Spider):
    name = "boligsidensellingspider"
    start_urls = ["https://www.boligsiden.dk/landsdel"]

    def parse(self, response):
        
        regions = response.css('a.flex.flex-row.items-center::attr(href)').getall()

        for region in regions:

            yield response.follow(url = region, callback= self.get_sellings)
    
    def get_sellings(self, response):

        link = response.url + '/tilsalg/villa,ejerlejlighed,raekkehus,villalejlighed,andelsbolig,landejendom'

        yield scrapy.Request(url = link, callback = self.parse_properties )

    def parse_properties(self, response):

        properties = response.css('div.overflow-hidden a.text-center::attr(href)').getall()

        for property in properties:
            
            yield scrapy.Request(url = 'https://www.boligsiden.dk' + property, callback = self.parse_property)

        next_page_url = response.css('a[rel="next"][aria-disabled="false"]::attr(href)').get()

        yield scrapy.Request(url = 'https://www.boligsiden.dk' + next_page_url, callback = self.parse_properties)
    
    def parse_property(self, response) :

        for script_text in response.css("script::text").getall():

            if 'props' in script_text:
                
                property_info = {}
                
                property_data = "".join(script_text)

                property_data = json.loads(property_data)

                property_data = property_data['props']['pageProps']['address']['case']

                property_info['source'] = response.url

                property_info['address_country'] = 'Denmark'
                
                property_info['property_listing_type'] = 'sale'

                try:
                    property_info['property_housing_type'] = property_data['addressType']
                except:
                    pass
                
                try:
                    property_info['timeline_time_on_market'] = property_data['timeOnMarket']
                except:
                    pass

                try:
                    property_info['latitude'] = property_data['coordinates']['lat']
                    property_info['longitude'] = property_data['coordinates']['lon']
                except:
                    pass
                
                try:
                    property_info['property_description'] = property_data['descriptionBody']
                except:
                    pass

                try:
                    property_info['property_energy_label'] = property_data['energyLabel']
                    # Function to separate letter and year
                    def separate_letter_year(value):
                        if isinstance(value, str):
                            match = re.match(r'([A-H])(\d{0,4})?', value)
                            if match:
                                letter = match.group(1)
                                year = match.group(2) if match.group(2) else None
                                return letter, year
                        return None, None

                    # Apply separation function to 'property_energy_label' value in property_info
                    if 'property_energy_label' in property_info:
                        letter, year = separate_letter_year(property_info['property_energy_label'])
                        property_info['property_energy_label'] = letter  # Replace with the letter only
                        if year is not None:
                            property_info['property_energy_label_year'] = int(year)  # Add year if present
                except:
                    pass

                try:
                    property_info['property_living_area_sqm'] = property_data['housingArea']
                except:
                    pass
                area = property_info["property_living_area_sqm"]
                if not area:
                    property_info['bucket'] = np.NaN
                elif area < 25:
                    property_info['bucket'] = 1
                elif area < 50:
                    property_info['bucket'] = 2
                elif area < 75:
                    property_info['bucket'] = 3
                elif area < 100:
                    property_info['bucket'] = 4
                elif area < 150:
                    property_info['bucket'] = 5
                else:
                    property_info['bucket'] = 6
                try:
                    property_info['property_plot_area_sqm'] = property_data['lotArea']
                except:
                    pass

                try:
                    property_info['property_monthly_expense'] = property_data['monthlyExpense']
                except:
                    pass

                try:
                    property_info['property_floor_number'] = property_data['numberOfFloors']
                except:
                    pass

                try:
                    property_info['property_rooms_amount'] = property_data['numberOfRooms']
                except:
                    pass

                try:
                    property_info['property_toilets_amount'] = property_data['numberOfToilets']
                except:
                    pass

                try:
                    property_info['property_sale_price_per_sqm'] = property_data['perAreaPrice']
                except:
                    pass

                try:
                    property_info['property_sale_price'] = property_data['priceCash']
                except:
                    pass

                property_info['property_sale_price_currency'] = 'DKK'
                
                try:
                    property_info['property_price_change_percentage'] = property_data['priceChangePercentage']
                except:
                    pass

                try:
                    property_info['property_weighted_area_sqm'] = property_data['weightedArea']
                except:
                    pass

                try:
                    property_info['property_construction_year'] = property_data['yearBuilt']
                except:
                    pass

                try:
                    property_info['property_buildings_info'] = property_data['address']['buildings']
                except:
                    pass

                try:
                    property_info['address_city'] = property_data['address']['cityName']
                except:
                    pass

                try:
                    house_number = property_data['address']['houseNumber']

                    match = re.match(r'(\d+)(.*)', house_number)
                    
                    if match:
                        property_info['address_house_number'] = match.group(1)
                        property_info['address_addition'] = match.group(2).strip() or None
                    else:
                        property_info['address_house_number'] = house_number

                except:
                    pass

                try:
                    property_info['property_latest_valuation'] = property_data['address']['latestValuation']
                except:
                    pass

                try:
                    property_info['address_municipality_info'] = property_data['address']['municipality']
                except:
                    pass

                try:
                    property_info['address_municipality'] = property_data['address']['municipality']['name']

                except:
                    pass

                try:
                    property_info['address_mun_code'] = property_data['address']['municipality']['code']
                
                except:
                    pass

                try:
                    property_info['address_place'] = property_data['address']['placeName']
                except:
                    pass

                try:
                    property_info['address_street'] = property_data['address']['roadName'] 
                except:
                    pass

                try:
                    property_info['address_postal_code'] = str(property_data['address']['zipCode'])
                
                except:
                    pass

                try:
                    property_info['media_set'] = [source['url'] 
                                            for property_item in property_data['images'] 
                                            for source in property_item['imageSources'] 
                                            if source['size']['height'] == 600]

                except:
                    pass
                yield property_info

                break




